#ifndef __PSPICEDIGAPIDEFS_H__
#define __PSPICEDIGAPIDEFS_H__

#include "PSpiceCommonAPIDefs.h"

#define UNSPEC 1e-33
#define PSP_VALUE_NOT_DEFINED(x) (x>=MAXREAL?true:false)

enum PSPICE_PORT_TYPE { PSPICE_PORT_IO = 0, PSPICE_PORT_IN = 1 };


//Following functions are provided by PSpice Engine:
PSP_CMI_EXPORT void PSpiceSetDelay(void* pRef, const PSpiceDelay* pDelay);
PSP_CMI_EXPORT void PSpiceSetConstraint(void* pRef, const PSpiceConstraint* pConstraint);
PSP_CMI_EXPORT void PSpiceSetOutputSpec(void* pRef, const PSpiceOutputSpec* pOutSpec);
PSP_CMI_EXPORT void PSpiceSetInputSpec(void* pRef, const PSpiceInputSpec* pInSpec);
PSP_CMI_EXPORT void PSpiceGetOutputSpec(void* pRef, PSpiceOutputSpec& pOutputSpec);
PSP_CMI_EXPORT void PSpiceGetInputSpec(void* pRef, PSpiceInputSpec& pInputSpec);
PSP_CMI_EXPORT bool PSpiceGetTimingModelValue(void* pRef, const char* pParamName, PSpiceDelay& pDelay);
PSP_CMI_EXPORT void PSpiceSetState(void* pRef, int pPortIndex, PSpiceState* pState, const PSpiceDelay* pDelay);	//chpDigitalC
PSP_CMI_EXPORT void PSpiceGetParameterValue(void* pRef, PSpiceCMIParam* pParam);	//chpDigitalC



/**
 *\n Templates for API's that need to be exposed by the Model Dll - their names follow the convention p*_t
 *\n    Their pointers will be set by the Model Dll into PSpice Engine using the descSet* Functions.
*/
/**
* @param	Instance Name
* @return	Digital Descriptor Reference
*/
typedef void* (*pCreateDevice_t)(const char* pInstName, void* ref);

/**
* @param	pRef	Instance Reference
* @return	None
*/
typedef void (*pInitDevice_t)(void* pRef);

/**
* @param	pRef		Instance Reference
* @Param	pTicks		TICKS
* @return	Status		false=>Failure true=>Success
*/
typedef bool(*pPSpiceEvaluateDevice_t)(void* pRef, double pTicks, PSpiceState* pVectorStates, int pSize); //chpDigitalC

/**
* @param	pRef		Instance Reference
* @param	pPortIndex	0-based port index on which to set state
* @param	pState		New State
* @param	pDelay		Delay to be used before setting the new state
*/
typedef void(*pPSpiceSetState_t)(void* pRef, int pPortIndex, PSpiceState* pState, const PSpiceDelay* pDelay); //chpDigitalC

/**
* @param	pRef		Instance Reference
* @param	pParam		PSpice Parameter
*/
typedef void(*pPSpiceGetParameterValue_t)(void* pRef, PSpiceCMIParam* pParam);

/**
* @param	pRef	Instance Reference
*/
typedef void(*pDeleteDevice_t)(void* pRef);

/**
* @param	Instance Reference
* @return	Number of terminals
*/
typedef int(*pGetDeviceTermCount_t) (void* pRef);

/**
* @return	true => Check PASS, else FAIL
*/
typedef bool(*pSetDeviceTermCount_t) (void* pRef, int pTermCount);
/**
* Get Term values from Model Dll to PSpice
* @param	pRef			Instance Reference
* @param	pVectorStates	Vector of values
*/
typedef int(*pGetDeviceTermValue_t)(void* pRef, PSpiceState* pVectorStates, int pSize);

/**
* Set Term values from PSpice to Model Dll
* @param	pRef			Instance Reference
* @param	pVectorStates	Vector of values
*/
typedef int(*pSetDeviceTermValue_t)(void* pRef, PSpiceState* pVectorStates, int pSize);

/**
* @param	pRef			Instance Reference
* @param	pVectorPorts	Vector of values
*/
typedef void(*pGetDeviceTermTypes_t)(void* pRef, void* pVectorPorts);

/**
* @param	pRef				Instance Reference
* @param	pParamCount			Length of vector
* @param	pVectorParameter	Vector of values
*/
typedef bool(*pSetParameter_t)(void* pRef, int pParamCount, void** pVectorParameter);

typedef const char* (*pDigPrintDescription_t) (void* pRef);

/**
* @return Ticks
*/
typedef double(*pGetTicksFromTime_t)(double pTime);

/**
* @return Time
*/
typedef double(*pGetTimeFromTicks_t)(double pTicks);


//pChanged = 0 => CHANGED
// = 1 => CHANGED_LH
// = 2 => CHANGED_HL
typedef bool(*pPSpiceChanged_t)(void* pRef, const char* pNetName, double pTime, int pChanged);

//pOut = output port name
// ptransition = TRN_LH | TRN_HL | ...
typedef bool(*pPSpiceGetTransition_t)(void* pRef, const char* pOut, const char* pTransition);

/**
* Function templates to be used by model dll to set and call functions in PSpice
*/

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to create Device function
*/
typedef void(*descSetCreateDevice_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to delete Device function
*/
typedef void(*descSetDeleteDevice_t)(void* pRef, void* pFuncPtr);

/**
* This function call is optional
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to get device terminal count function
*/
typedef void(*descSetGetDeviceTermCount_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to initialize Device function
*/
typedef void(*descSetInitDevice_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to set Device terminal count function
*/
typedef void(*descSetSetDeviceTermCount_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to evaluate Device function
*/
typedef void(*descSetEvaluateDevice_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to get Device Terminal value function
*/
typedef void(*descSetGetDeviceTermValue_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to set device terminal value function
*/
typedef void(*descSetSetDeviceTermValue_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to get device terminals function
*/
typedef void(*descSetGetDeviceTerminals_t)(void* pRef, void* pFuncPtr);

/**
* @param	pRef		Instance Reference
* @param	pFuncPtr	Pointer to set Parameter function
*/
typedef void(*descSetSetParameter_t)(void* pRef, void* pFuncPtr);


/**
* Set Delay - optional function - the same functionality is also available in setState
* enum value = PSPICE_SET_DELAY
* @param	pRef	Model reference
* @param	pDelay	Delay pointer
* @return	None
*/
typedef void(*pPSpiceSetDelay_t)(void* pRef, const PSpiceDelay* pDelay);

/**
* Set Constraint
* enum value = PSPICE_SET_CONSTRAINT
* @param	pRef	Model reference
* @param	pDelay	Delay pointer
* @return	None
*/
typedef void(*pPSpiceSetConstraint_t)(void* pRef, const PSpiceConstraint* pConstraint);

/**
* Following functions should be called before evaluate - they make changes on all terminals of the device
*/
/**
* Setup Input Buffer specification
* enum value = PSPICE_SET_INSPEC
* @param	pRef		Model reference
* @param	pInSpec		Input specification
* @return	None
*/
typedef void(*pPSpiceSetInputSpec_t)(void* pRef, const PSpiceInputSpec* pInSpec);

/**
* Setup Output Buffer Specification
* enum value = PSPICE_SET_OUTSPEC
* @param	pRef		Model reference
* @param	pOutSpec	Output specification
* @return	None
*/
typedef void(*pPSpiceSetOutputSpec_t)(void* pRef, PSpiceOutputSpec* pOutSpec);

/**
* Get Input Buffer Specification
* enum value = PSPICE_GET_INSPEC
* @param	pRef		Model reference
* @param	pInSpec	Input specification
* @return	None
*/
typedef void(*pPSpiceGetInputSpec_t)(void* pRef, PSpiceInputSpec& pInSpec);

/**
* Get Output Buffer Specification
* enum value = PSPICE_GET_OUTSPEC
* @param	pRef		Model reference
* @param	pOutSpec	Output specification
* @return	None
*/
typedef void(*pPSpiceGetOutputSpec_t)(void* pRef, const PSpiceOutputSpec& pOutSpec);

/**
* Get Timing Model Value
* enum value = PSPICE_GET_TIMINGVALUE
* @param	pRef		Model reference
* @param	pParamName	Parameter Name
* @param	pDelay		Delay spec
* @return	None
*/
typedef bool(*pPSpiceGetTimingModelValue_t)(void* pRef, const char* pParamName, PSpiceDelay& pDelay);

class PSpiceNetsList {
public:
	char mNetName[1024];
	PSpiceNetsList* mNext;

	PSpiceNetsList() {
		mNext = NULL;
	}
};

class PSpiceSetupHoldConstraint {
public:
	bool mSetupHoldSpecified; /* Constraint is specified by the user - should be used by simulator */

	char mClockName[1024];
	PSpiceNetsList* mNetsList;
	int mCountData;
	float setuptime_lo;   /* time data must be stable lo before clock assert */
	float setuptime_hi;   /* time data must be stable hi before clock assert */
	float holdtime_lo;    /* time data must be stable lo after clock assert */
	float holdtime_hi;    /* time data must be stable hi after clock assert */
	float releasetime_lh; /* time data must undergo lh before clk assert */
	float releasetime_hl; /* time data must undergo hl before clk assert */

	//	struct vio_list* vio_list_p;        /* list of violations associated with clk chg */
	bool clk_assertion; // 0 => HL, 1=>LH                /* The assertion type of the CLOCK; l->H, h->L */
	/* NOTE: only the "final" level (H,L) is kept. */

	PSpiceSetupHoldConstraint() {
		mSetupHoldSpecified = false;
		mNetsList = new PSpiceNetsList();

		setuptime_lo = (float)0.0;
		setuptime_hi = setuptime_lo;

		releasetime_lh = (float)0.0;
		releasetime_hl = releasetime_lh;

		holdtime_lo = (float)0.0;
		holdtime_hi = holdtime_lo;

		clk_assertion = 0;
	}
	~PSpiceSetupHoldConstraint() {
		delete mNetsList;
		mNetsList = NULL;
	}
};

/**
* Class for Width Constraint Definition
*/
class PSpiceWidthConstraint {
public:
	bool mWidthSpecified; /* Constraint is specified by the user - should be used by simulator */

	/**
	* Name of input being checked
	*/
	char mInputNode[1024];

	/**
	* Minimum time that input must remain in H state
	*/
	float min_high;

	/**
	* Minimum time that input must remain in L state
	*/
	float min_low;

	PSpiceWidthConstraint() {
		mWidthSpecified = false;
	}
};

/**
* Class for Frequency Constraint Definition
*/
class PSpiceFreqConstraint {
public:
	bool mFreqSpecified; /* Constraint is specified by the user - should be used by simulator */

	/**
	* Name of input being checked
	*/
	char* mInputNode;

	/**
	* Minimum frequency (Hz)
	*/
	float min_freq;

	/**
	* Maximum frequency (Hz)
	*/
	float max_freq;

	char errorflags;

	PSpiceFreqConstraint() {
		mFreqSpecified = false;
		min_freq = 0;
		max_freq = 0;
		errorflags = 0;
	}
};

/**
* Composite class for definition of all constraints
*/
class PSpiceConstraint {
public:
	PSpiceSetupHoldConstraint mSetupHold;
	PSpiceWidthConstraint mWidth;
	PSpiceFreqConstraint mFreq;
};

class PSpiceDelay {
public:
	double mMinDelay;
	double mTypDelay;
	double mMaxDelay;
};

#define MAXIOLEVEL 4       // number of IO Levels in UIO model

/** Output Buffer Specification. */
class PSpiceOutputSpec {
public:
	double l_drive;		/**< drive R at low */
	double h_drive;		/**< drive R at high */
	double z_drive;				/**< leakage R for Z state */
	double load;				/**< capacitive load */
	double tswhl[MAXIOLEVEL];	/**< switching time - high to low */
	double tswlh[MAXIOLEVEL];	/**< switching time - low to high */
	double pwrt;		/**< pulse width rejection threshold */

	PSpiceOutputSpec() {
		l_drive = UNSPEC;
		h_drive = UNSPEC;
		z_drive = UNSPEC;
		load = UNSPEC;
		for (int i = 0; i<MAXIOLEVEL; i++) {
			tswhl[i] = UNSPEC;
			tswlh[i] = UNSPEC;
		}
		pwrt = UNSPEC;
	}
};

/** Input Buffer Specification. */
class PSpiceInputSpec {
public:
	double load;    /**< capacitive load */
	double inR;     /**< input Resistive load */
	double Tstore;  /**< minimum storage time for chrg ctrl net */
};

/** PSpice Port */
class PSpicePort {
public:
	const char* mName;
	// number is needed for instantiation - in engine i assume that i am getting the correct numbers
	int mNumber;
	int mType;
};

/** Digital State */
class PSpiceState {
public:
	union {
		struct {
			unsigned level : 3;
			unsigned str0 : 6;
			unsigned str1 : 6;
			unsigned _filler : 1;

			// upper half is Hazard "key"
			unsigned hazardtype : 3;
			unsigned multiple : 1;
			unsigned msgid : 10;
			unsigned notposted : 1;
			unsigned persistent : 1;
		} fields;
		unsigned long stateVal;
	} val;
	inline bool isZ() const {
		return ((val.stateVal & 0x0000FFF8) == 0);
	}
	inline int getLevel() const   {
		return val.fields.level;
	}
	inline PSpiceState& operator = (const int pBit)  {
		val.fields.level = pBit;
		return *this;
		if (pBit<5){
			val.fields.level = pBit;
		}
		else {
			val.fields.level = (unsigned char)2;
			val.fields.str0 = 0;
			val.fields.str1 = 0;
		}
		return *this;
	}
};


class pspBit;

static const char AND_TABLE[6][6] =
{   // 0,  1,  X,  R,  F,  Z
	{ '0', '0', '0', '0', '0', '0' } //0
	, { '0', '1', 'X', 'X', 'X', 'X' } //1
	, { '0', 'X', 'X', 'X', 'X', 'X' } //X
	, { '0', 'X', 'X', 'X', 'X', 'X' } //R
	, { '0', 'X', 'X', 'X', 'X', 'X' } //F
	, { '0', 'X', 'X', 'X', 'X', 'X' } //Z
};

static const char OR_TABLE[6][6] =
{   // 0,  1,  X,  R,  F,  Z
	{ '0', '1', 'X', 'X', 'X', 'X' } //0
	, { '1', '1', '1', '1', '1', '1' } //1
	, { 'X', '1', 'X', 'X', 'X', 'X' } //X
	, { 'X', '1', 'X', 'X', 'X', 'X' } //R
	, { 'X', '1', 'X', 'X', 'X', 'X' } //F
	, { 'X', '1', 'X', 'X', 'X', 'X' } //Z
};

static const char XOR_TABLE[6][6] =
{   // 0,  1,  X,  R,  F,  Z
	{ '0', '1', 'X', 'X', 'X', 'X' } //0
	, { '1', '0', 'X', 'X', 'X', 'X' } //1
	, { 'X', 'X', 'X', 'X', 'X', 'X' } //X
	, { 'X', 'X', 'X', 'X', 'X', 'X' } //R
	, { 'X', 'X', 'X', 'X', 'X', 'X' } //F
	, { 'X', 'X', 'X', 'X', 'X', 'X' } //Z
};

static const char NOT_TABLE[6] =
{   // 0,  1,  X,  R,  F,  Z
	'1', '0', 'X', 'F', 'R', 'X'
};

class pspBit {

public:

	enum pspBitLevels {
		LO = 0,
		HI = 1,
		UNKNOWN = 2,
		RISE = 3,
		FALL = 4,
		HIZ = 5
	};

	pspBit(char pChar) {
		mValue = pChar;
	}

	pspBit(PSpiceState &pState) {
		mValue = "01XRFZ"[((long)(pState.val.stateVal) & (long)0x00000007)];
	}

	pspBit(bool pBool = false) {
		mValue = pBool ? '1' : '0';
	}

	inline operator char() const {
		return mValue;
	}

	inline operator int() const {
		switch (mValue) {
		case '0': return 0;
		case '1': return 1;
		case 'X': return 2;
		case 'R': return 3;
		case 'F': return 4;
		case 'Z': return 5;
		};
		return mValue;
	}

	inline pspBit& operator = (const pspBit& pBit)  {
		mValue = pBit.mValue; return *this;
	}

	inline pspBit& operator = (const int pValue) {
		switch (pValue) {
		case 0:
			mValue = '0';
			break;
		case 1:
			mValue = '1';
			break;
		case 2:
			mValue = 'X';
			break;
		case 3:
			mValue = 'R';
			break;
		case 4:
			mValue = 'F';
			break;
		case 5:
			mValue = 'Z';
			break;
		};
		return *this;
	}

	friend bool operator == (const pspBit& pBit1, const pspBit& pBit2);
	friend pspBit operator | (const pspBit& pBit1, const pspBit& pBit2);
	friend pspBit operator & (const pspBit& pBit1, const pspBit& pBit2);
	friend inline pspBit operator ~ (const pspBit& pBit);
	friend pspBit operator ^ (const pspBit& pBit1, const pspBit& pBit2);

private:
	char mValue;
};

inline bool operator == (const pspBit& pBit1, const pspBit& pBit2){
	return (pBit1.mValue == pBit2.mValue ? true : false);
}

inline pspBit operator ^ (const pspBit& pBit1, const pspBit& pBit2) {
	char lValue = XOR_TABLE[(int)pBit1][(int)pBit2];
	return pspBit(lValue);
}

inline pspBit operator & (const pspBit& pBit1, const pspBit& pBit2) {
	char lValue = AND_TABLE[(int)pBit1][(int)pBit2];
	return pspBit(lValue);
}

inline pspBit operator | (const pspBit& pBit1, const pspBit& pBit2){
	char lValue = OR_TABLE[(int)pBit1][(int)pBit2];
	return pspBit(lValue);
}

inline pspBit operator ~ (const pspBit& pBit) {
	char lValue = NOT_TABLE[(int)pBit];
	return pspBit(lValue);
}


#endif
