/***
*** Interface: pspice 16.6
 *** created by: admsXml-2.3.0 - Mon, 25 Jul 2016 14:36:59 ***/
/*
* Copyright by Cadence Design Systems 2014. All rights reserved
*/
#ifndef  __PSpiceBase_h__
#define  __PSpiceBase_h__
#include <string>
#include <math.h>
#include <cmath>
#include <map>
#include "pspiceCMIApiDefs.h"

#define MAXREAL 1e+30
#define MINREAL     1e-30
#define FALSE 0
#define OK 1

#define DLL_FUNC extern "C" __declspec(dllexport) // for C Files
typedef double* MTX_IDX;
typedef int CKT_IDX;
struct cmplex
{
public:
    double im;
    double re;
	cmplex(double pReal, double pImag) {
		re = pReal;
		im = pImag;
	}
};
// state vector
class qi_def
{
public:
    double qi_q;       /* charge on plates */
    double qi_i;       /* current thru capacitor */
};

//simulator uses last four values
#define MSTVCT 4

class PSpiceModel {
public:
	virtual void setParamValue(const char* pName, double pValue) {}
	virtual int ModChk() {return 0;}
	virtual void TmpMod(                 /* Update ONE model to a new temperature*/
		double Told,                     /* Temperature models have now          */
		double Tnew,                     /* Temperature models will have on exit */
		double tnom)                     /* nominal temperature                  */
		{}
	  bool mParamGivenInitialized;
	  PSpiceModel() {
		  mParamGivenInitialized = false;
	  }
};

class PSpiceDevice {
public:
	virtual void createMatrixStencil() {};
	virtual int ModChk() {return 0;};
	virtual int Load(int, int, int, double) {return 0;};
	virtual int Load(PSpiceDeviceMiscInfo* pInfo, int CKTmode) {return 0;}
	virtual int ACLoad(double) {return 0;};
	virtual double Noise(
        double tnow,		// present device temperature
        double Freq,		// frequency (Hz)
        double* ncmps,		// noise components...
		int pCompsLength) const {
			return 0.0;
		}
	virtual double Trunc() {return MAXREAL;};
	virtual void setParamValue(const char* pName, double pValue) {};
	virtual void setModel(PSpiceModel* pModel) {};
	virtual void setNodeName(int i, const char* pNodeName) {};
	virtual void LoadCheckpointInfo(FILE * fp, unsigned int pVersion) {};
	virtual void SaveCheckpointInfo(FILE * fp, unsigned int pVersion) {};
	virtual int AddInternalNodes(PSpiceDeviceMiscInfo* pMiscInfo) = 0;
	virtual double IPin(int pin, int mode) {return 0;};
	virtual cmplex IPinComplex(int pin, double omega) {return cmplex(0, 0);};
    virtual void TmpMod(
        double /* ToldK */,
        double /* TnewK */,
		double /* TnomK */) {}
	virtual size_t SaveTopology(FILE * fp,unsigned int pVersion) {return 0;};
	virtual bool CheckTopology(FILE * fp,unsigned int pVersion) {return true;};
	virtual size_t getTopologySize(unsigned int pVersion) {return 0;};
	// helper functions
	double* getRHS(int pNode) {
		std::map<int, double*>::iterator it;
		it = mResiduals.find(pNode);
		if(it == mResiduals.end()) {
			printf("\n ERROR: %s", pNode);
			return NULL;
		}
		return it->second;
	}
	virtual void initMatrix() {}
	  std::map < int, double* > mResiduals;
	  bool mParamGivenInitialized;
	  PSpiceDevice() {
		  mParamGivenInitialized = false;
	  }
};

const double CHARGE	= 1.6021918e-19;              // Charge on an electron
const double BOLTZ 	= 1.3806226e-23;              // Boltzmann's constant
const double CTOK	= 273.15e0;                   // 0 deg. C in deg. Kelvin
const double EPS0	= 8.854214871e-12;            // Permittivity of free space
const double EPSSIL	= 11.7e0*EPS0;                // Permittivity of silicon
const double EPSOX 	= 3.9e0*EPS0;                 // Permittivity of oxide
const double TWOPI 	= 8.0e0*atan2(1.0e0,1.0e0);   // Twice PI
const double RAD	= 360.0e0/TWOPI;              // Degrees per radian
const double LOG2	= log(2.0e0);                 // Log base e of 2
const double LOG10 	= log(10.0e0);                // Log base e of 10
const double ROOT2 	= sqrt(2.0e0);                // Square root of 2
const double NI		= 1.45e16;                    // Intrinsic concentration
const double PI		= 4.0e0*atan2(1.0e0,1.0e0);   // BSim4:Intrinsic concentration


inline void UPPERLIM(double& e, double l) 
{
    // UPPERLIM(e,l) will force e to be <= l.
    if (e  >  double(l))  e = double(l);
};

inline void LOWERLIM(double& e, double l) 
{
    // LOWERLIM(e,l) will force e to be >= l.
    if (e  <  double(l))  e = double(l);
};


const double AF_LIMIT = double (0.1);
const double EG_LIMIT = double (0.1);
const double FC_LIMIT = double (0.9999);
inline void RG(double& v) 
{
    // RG inverts the value of v unless it is 0.0.
    if (v != double(0.0))  v = double(1.0 / v);
};
#define YES 1
#define NO 0
#define PNJCHG(chg,cap,cjo,vj,pb,fcpb,m,f1,f2,f3)\
    if ( (pb) == 0.) {\
      printf("PNJCHG:  divide by zero in %s", mName); \
      }\
    else if ( (vj) < (fcpb) ) {\
      double\
        arg  = 1-(vj)/(pb),\
        sarg = (m)==.5 ? 1/sqrt(arg) : pow(arg,-(m));\
      chg = (cjo)*(pb)*( (m)==1.? -log(arg) : (1-arg*sarg)/(1-(m)) );\
      cap = (cjo)*sarg;\
      }\
    else {\
      chg = (cjo)*( (f1) + ( (f3)*((vj)-(fcpb)) + ( (m)/((pb)+(pb)) )*( (vj)*(vj)-(fcpb)*(fcpb) ) )/(f2) );\
      cap = (cjo)*( (f3) + (m)*(vj)/(pb) )/(f2);\
      }
/* same as PNJCHG, except FC==0 */
#define PNJCHG0(chg,cap,cjo,vj,pb,m)\
    if ( (pb) == 0.) {\
           printf("Error in PNJCHG0:  divide by zero in %s", GetDeviceName(Instance)); \
      }\
    else if ( (vj) < 0.) {\
      double\
        arg  = 1-(vj)/(pb),\
        sarg = (m)==.5 ? 1/sqrt(arg) : pow(arg,-(m));\
      chg = (cjo)*(pb)*( (m)==1.? -log(arg) : (1-arg*sarg)/(1-(m)) );\
      cap = (cjo)*sarg;\
      }\
    else {\
      chg = (vj)*(cjo)*(1 + (m)*(vj)/((pb)+(pb)));\
      cap =      (cjo)*(1 + (m)*(vj)/(pb));\
    }

/**
PNJLIM expansion (using regular logarithms, not the MIN(log(x),10.) )
*/
inline int PNJLIM(double& Vnew,double Vold,double Vte, double Vcrit)
{
    if ( Vnew <= Vcrit || Vnew <= 0.0 || fabs(Vnew - Vold) <= Vte + Vte ) {
        return NO;
    }
    if ( Vold <= 0. ) {
        Vnew = Vte * log( Vnew / Vte );
    }
    else if ( Vnew <= Vold - Vte ) {
        Vnew = Vcrit;
    }
    else {
        Vnew = Vold + Vte * log( 1 + (Vnew-Vold) / Vte );
    }
    return YES;
}
const double OMIT = (1.10250000000000000E-36);
#define MAX(A,B)    ((A) > (B) ? (A) : (B))
#define MIN(A,B)    ((A) < (B) ? (A) : (B))
#define EXP(a) exp(a)          /*( (a < -50.) ? 0.0 : exp(MIN(a,50.)) )*/
#define LOG(a) Ln(a)            /*(a < 1.928752e-22) ? -50. : log(a) )*/
#define NOSOLV false
const double MIN_INOISE = 1e-20;   /* lower limit of device noise currents */
inline double SUM_SQR(double x, double y) 
{
    return x*x + y*y;
}
inline cmplex operator-( const cmplex & a )
{
    return( cmplex( -a.re, -a.im ) ) ;
}
inline cmplex operator*( const cmplex & a, const cmplex & b )
{
	double realPart      = a.re * b.re - a.im * b.im ;
	double imaginaryPart = a.re * b.im + a.im * b.re ;
	return( cmplex( realPart, imaginaryPart ) ) ;
}

#endif
